import time

from paasta_tools.deployd.common import PaastaThread


class QueueMetrics(PaastaThread):
    def __init__(self, inbox, cluster, metrics_provider):
        super().__init__()
        self.daemon = True
        self.metrics = metrics_provider

        self.inbox = inbox.to_bounce
        self.instances_to_bounce_later = inbox.instances_to_bounce_later
        self.instances_to_bounce_now = inbox.instances_to_bounce_now

        self.instances_to_bounce_later_gauge = self.metrics.create_gauge(
            "instances_to_bounce_later", paasta_cluster=cluster
        )
        self.instances_that_need_to_be_checked_up_on_gauge = self.metrics.create_gauge(
            "instances_that_need_to_be_checked_up_on", paasta_cluster=cluster
        )
        self.instances_to_bounce_now_gauge = self.metrics.create_gauge(
            "instances_to_bounce_now", paasta_cluster=cluster
        )

    def run(self):
        while True:
            self.instances_to_bounce_later_gauge.set(
                self.instances_to_bounce_later.qsize()
            )
            self.instances_that_need_to_be_checked_up_on_gauge.set(
                len(self.inbox.keys())
            )
            self.instances_to_bounce_now_gauge.set(self.instances_to_bounce_now.qsize())
            time.sleep(20)



import copy
import json
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from kazoo.client import KazooClient
from kazoo.exceptions import BadVersionError
from kazoo.exceptions import NodeExistsError
from kazoo.exceptions import NoNodeError
from kazoo.protocol.states import ZnodeStat

from paasta_tools.utils import _log


class MesosTaskParametersIsImmutableError(Exception):
    pass


_SelfT = TypeVar("_SelfT", bound="MesosTaskParameters")


class MesosTaskParameters:
    health: Any
    mesos_task_state: str
    is_draining: bool
    is_healthy: bool
    offer: Any
    resources: Any

    def __init__(
        self,
        health=None,
        mesos_task_state=None,
        is_draining=None,
        is_healthy=None,
        offer=None,
        resources=None,
    ):
        self.__dict__["health"] = health
        self.__dict__["mesos_task_state"] = mesos_task_state
        self.__dict__["is_draining"] = is_draining
        self.__dict__["is_healthy"] = is_healthy
        self.__dict__["offer"] = offer
        self.__dict__["resources"] = resources

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return "{}(\n    {})".format(
            type(self).__name__,
            ",\n    ".join(["%s=%r" % kv for kv in self.__dict__.items()]),
        )

    def __setattr__(self, name, value):
        raise MesosTaskParametersIsImmutableError()

    def __delattr__(self, name):
        raise MesosTaskParametersIsImmutableError()

    def merge(self: _SelfT, **kwargs) -> "MesosTaskParameters":
        """Return a merged MesosTaskParameters object, where attributes in other take precedence over self."""

        new_dict = copy.deepcopy(self.__dict__)
        new_dict.update(kwargs)

        return MesosTaskParameters(**new_dict)

    @classmethod
    def deserialize(cls: Type[_SelfT], serialized_params: Union[str, bytes]) -> _SelfT:
        return cls(**json.loads(serialized_params))

    def serialize(self):
        return json.dumps(self.__dict__).encode("utf-8")


class TaskStore:
    def __init__(self, service_name, instance_name, framework_id, system_paasta_config):
        self.service_name = service_name
        self.instance_name = instance_name
        self.framework_id = framework_id
        self.system_paasta_config = system_paasta_config

    def get_task(self, task_id: str) -> MesosTaskParameters:
        """Get task data for task_id. If we don't know about task_id, return None"""
        raise NotImplementedError()

    def get_all_tasks(self) -> Dict[str, MesosTaskParameters]:
        """Returns a dictionary of task_id -> MesosTaskParameters for all known tasks."""
        raise NotImplementedError()

    def overwrite_task(self, task_id: str, params: MesosTaskParameters) -> None:
        raise NotImplementedError()

    def add_task_if_doesnt_exist(self, task_id: str, **kwargs) -> None:
        """Add a task if it does not already exist. If it already exists, do nothing."""
        if self.get_task(task_id) is not None:
            return
        else:
            self.overwrite_task(task_id, MesosTaskParameters(**kwargs))

    def update_task(self, task_id: str, **kwargs) -> MesosTaskParameters:
        existing_task = self.get_task(task_id)
        if existing_task:
            merged_params = existing_task.merge(**kwargs)
        else:
            merged_params = MesosTaskParameters(**kwargs)

        self.overwrite_task(task_id, merged_params)
        return merged_params

    def garbage_collect_old_tasks(self, max_dead_task_age: float) -> None:
        # TODO: call me.
        # TODO: implement in base class.
        raise NotImplementedError()

    def close(self):
        pass


class DictTaskStore(TaskStore):
    def __init__(self, service_name, instance_name, framework_id, system_paasta_config):
        self.tasks: Dict[str, MesosTaskParameters] = {}
        super().__init__(
            service_name, instance_name, framework_id, system_paasta_config
        )

    def get_task(self, task_id: str) -> MesosTaskParameters:
        return self.tasks.get(task_id)

    def get_all_tasks(self) -> Dict[str, MesosTaskParameters]:
        """Returns a dictionary of task_id -> MesosTaskParameters for all known tasks."""
        return dict(self.tasks)

    def overwrite_task(self, task_id: str, params: MesosTaskParameters) -> None:
        # serialize/deserialize to make sure the returned values are the same format as ZKTaskStore.
        self.tasks[task_id] = MesosTaskParameters.deserialize(params.serialize())


class ZKTaskStore(TaskStore):
    def __init__(self, service_name, instance_name, framework_id, system_paasta_config):
        super().__init__(
            service_name, instance_name, framework_id, system_paasta_config
        )
        self.zk_hosts = system_paasta_config.get_zk_hosts()

        # For some reason, I could not get the code suggested by this SO post to work to ensure_path on the chroot.
        # https://stackoverflow.com/a/32785625/25327
        # Plus, it just felt dirty to modify instance attributes of a running connection, especially given that
        # KazooClient.set_hosts() doesn't allow you to change the chroot. Must be for a good reason.

        chroot = f"task_store/{service_name}/{instance_name}/{framework_id}"

        temp_zk_client = KazooClient(hosts=self.zk_hosts)
        temp_zk_client.start()
        temp_zk_client.ensure_path(chroot)
        temp_zk_client.stop()
        temp_zk_client.close()

        self.zk_client = KazooClient(hosts=f"{self.zk_hosts}/{chroot}")
        self.zk_client.start()
        self.zk_client.ensure_path("/")

    def close(self):
        self.zk_client.stop()
        self.zk_client.close()

    def get_task(self, task_id: str) -> MesosTaskParameters:
        params, stat = self._get_task(task_id)
        return params

    def _get_task(self, task_id: str) -> Tuple[MesosTaskParameters, ZnodeStat]:
        """Like get_task, but also returns the ZnodeStat that self.zk_client.get() returns """
        try:
            data, stat = self.zk_client.get("/%s" % task_id)
            return MesosTaskParameters.deserialize(data), stat
        except NoNodeError:
            return None, None
        except json.decoder.JSONDecodeError:
            _log(
                service=self.service_name,
                instance=self.instance_name,
                level="debug",
                component="deploy",
                line=f"Warning: found non-json-decodable value in zookeeper for task {task_id}: {data}",
            )
            return None, None

    def get_all_tasks(self):
        all_tasks = {}

        for child_path in self.zk_client.get_children("/"):
            task_id = self._task_id_from_zk_path(child_path)
            params = self.get_task(task_id)
            # sometimes there are bogus child ZK nodes. Ignore them.
            if params is not None:
                all_tasks[task_id] = params

        return all_tasks

    def update_task(self, task_id: str, **kwargs):
        retry = True
        while retry:
            retry = False
            existing_task, stat = self._get_task(task_id)

            zk_path = self._zk_path_from_task_id(task_id)
            if existing_task:
                merged_params = existing_task.merge(**kwargs)
                try:
                    self.zk_client.set(
                        zk_path, merged_params.serialize(), version=stat.version
                    )
                except BadVersionError:
                    retry = True
            else:
                merged_params = MesosTaskParameters(**kwargs)
                try:
                    self.zk_client.create(zk_path, merged_params.serialize())
                except NodeExistsError:
                    retry = True

        return merged_params

    def overwrite_task(
        self, task_id: str, params: MesosTaskParameters, version=-1
    ) -> None:
        try:
            self.zk_client.set(
                self._zk_path_from_task_id(task_id), params.serialize(), version=version
            )
        except NoNodeError:
            self.zk_client.create(
                self._zk_path_from_task_id(task_id), params.serialize()
            )

    def _zk_path_from_task_id(self, task_id: str) -> str:
        return "/%s" % task_id

    def _task_id_from_zk_path(self, zk_path: str) -> str:
        return zk_path.lstrip("/")



import logging
import os
import time
from functools import reduce
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

import pyinotify
from kazoo.protocol.states import EventType
from kazoo.recipe.watchers import ChildrenWatch
from kazoo.recipe.watchers import DataWatch
from requests.exceptions import RequestException

from paasta_tools.deployd.common import get_marathon_clients_from_config
from paasta_tools.deployd.common import get_service_instances_needing_update
from paasta_tools.deployd.common import PaastaThread
from paasta_tools.deployd.common import rate_limit_instances
from paasta_tools.deployd.common import ServiceInstance
from paasta_tools.long_running_service_tools import AUTOSCALING_ZK_ROOT
from paasta_tools.marathon_tools import DEFAULT_SOA_DIR
from paasta_tools.marathon_tools import deformat_job_id
from paasta_tools.marathon_tools import get_marathon_apps_with_clients
from paasta_tools.mesos_maintenance import get_draining_hosts
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import list_all_instances_for_service
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import PATH_TO_SYSTEM_PAASTA_CONFIG_DIR


class PaastaWatcher(PaastaThread):
    def __init__(self, instances_to_bounce_later, cluster, config):
        super().__init__()
        self.daemon = True
        self.instances_to_bounce_later = instances_to_bounce_later
        self.cluster = cluster
        self.config = config
        self.is_ready = False


class AutoscalerWatcher(PaastaWatcher):
    def __init__(self, instances_to_bounce_later, cluster, config, **kwargs):
        super().__init__(instances_to_bounce_later, cluster, config)
        self.zk = kwargs.pop("zookeeper_client")
        self.watchers: Dict[str, PaastaWatcher] = {}

    def watch_folder(self, path, enqueue_children=False):
        """recursive nonsense"""
        if "autoscaling.lock" in path:
            return
        if path.split("/")[-1] == "instances":
            self.watch_node(path, enqueue=enqueue_children)
            return
        self.log.info(f"Adding folder watch on {path}")
        watcher = ChildrenWatch(
            self.zk, path, func=self.process_folder_event, send_event=True
        )
        self.watchers[path] = watcher
        children = watcher._client.get_children(watcher._path)
        if children:
            for child in children:
                self.watch_folder(f"{path}/{child}", enqueue_children=enqueue_children)

    def _enqueue_service_instance(self, path):
        service, instance = path.split("/")[-3:-1]
        self.log.info(
            f"Number of instances changed for {service}.{instance} by the autoscaler."
        )
        # https://github.com/python/mypy/issues/2852
        service_instance = ServiceInstance(  # type: ignore
            service=service,
            instance=instance,
            cluster=self.config.get_cluster(),
            bounce_by=int(time.time()),
            bounce_timers=None,
            watcher=type(self).__name__,
            failures=0,
        )
        self.instances_to_bounce_later.put(service_instance)

    def watch_node(self, path, enqueue=False):
        self.log.info(f"Adding zk node watch on {path}")
        DataWatch(self.zk, path, func=self.process_node_event, send_event=True)
        if enqueue:
            self._enqueue_service_instance(path)

    def process_node_event(self, data, stat, event):
        self.log.debug(f"zk node change: {event}")
        if event and (
            event.type == EventType.CREATED or event.type == EventType.CHANGED
        ):
            self._enqueue_service_instance(event.path)

    def process_folder_event(self, children, event):
        self.log.debug(f"Folder change: {event}")
        if event and (event.type == EventType.CHILD):
            fq_children = [f"{event.path}/{child}" for child in children]
            for child in fq_children:
                if child not in self.watchers:
                    self.watch_folder(child, enqueue_children=True)

    def run(self):
        if not self.zk.exists(AUTOSCALING_ZK_ROOT):
            self.zk.ensure_path(AUTOSCALING_ZK_ROOT)
        self.watchers[AUTOSCALING_ZK_ROOT] = self.watch_folder(AUTOSCALING_ZK_ROOT)
        self.is_ready = True
        while True:
            time.sleep(0.1)


class SoaFileWatcher(PaastaWatcher):
    def __init__(self, instances_to_bounce_later, cluster, config, **kwargs):
        super().__init__(instances_to_bounce_later, cluster, config)
        self.wm = pyinotify.WatchManager()
        self.wm.add_watch(DEFAULT_SOA_DIR, self.mask, rec=True)
        self.notifier = pyinotify.Notifier(
            watch_manager=self.wm,
            default_proc_fun=YelpSoaEventHandler(filewatcher=self),
        )

    @property
    def mask(self):
        boring_flags = ["IN_CLOSE_NOWRITE", "IN_OPEN", "IN_ACCESS", "IN_ATTRIB"]
        return reduce(
            lambda x, y: x | y,
            [
                v
                for k, v in pyinotify.EventsCodes.OP_FLAGS.items()
                if k not in boring_flags
            ],
        )

    def run(self):
        self.notifier.loop(callback=self.startup_checker)

    def startup_checker(self, obj):
        self.is_ready = True


class PublicConfigFileWatcher(PaastaWatcher):
    def __init__(self, instances_to_bounce_later, cluster, config, **kwargs):
        super().__init__(instances_to_bounce_later, cluster, config)
        self.wm = pyinotify.WatchManager()
        self.wm.add_watch(PATH_TO_SYSTEM_PAASTA_CONFIG_DIR, self.mask, rec=True)
        self.notifier = pyinotify.Notifier(
            watch_manager=self.wm,
            default_proc_fun=PublicConfigEventHandler(filewatcher=self),
        )

    @property
    def mask(self):
        boring_flags = ["IN_CLOSE_NOWRITE", "IN_OPEN", "IN_ACCESS", "IN_ATTRIB"]
        return reduce(
            lambda x, y: x | y,
            [
                v
                for k, v in pyinotify.EventsCodes.OP_FLAGS.items()
                if k not in boring_flags
            ],
        )

    def run(self):
        self.notifier.loop(callback=self.startup_checker)

    def startup_checker(self, obj):
        self.is_ready = True


class MaintenanceWatcher(PaastaWatcher):
    def __init__(self, instances_to_bounce_later, cluster, config, **kwargs):
        super().__init__(instances_to_bounce_later, cluster, config)
        self.draining: Set[str] = set()
        self.marathon_clients = get_marathon_clients_from_config()

    def get_new_draining_hosts(self):
        try:
            draining_hosts = get_draining_hosts()
        except RequestException as e:
            self.log.error(f"Unable to get list of draining hosts from mesos: {e}")
            draining_hosts = list(self.draining)
        new_draining_hosts = [
            host for host in draining_hosts if host not in self.draining
        ]
        for host in new_draining_hosts:
            self.draining.add(host)
        hosts_finished_draining = [
            host for host in self.draining if host not in draining_hosts
        ]
        for host in hosts_finished_draining:
            self.draining.remove(host)
        return new_draining_hosts

    def run(self):
        self.is_ready = True
        while True:
            new_draining_hosts = self.get_new_draining_hosts()
            service_instances: List[ServiceInstance] = []
            if new_draining_hosts:
                self.log.info(f"Found new draining hosts: {new_draining_hosts}")
                service_instances = self.get_at_risk_service_instances(
                    new_draining_hosts
                )
            for service_instance in service_instances:
                self.instances_to_bounce_later.put(service_instance)
            time.sleep(self.config.get_deployd_maintenance_polling_frequency())

    def get_at_risk_service_instances(self, draining_hosts) -> List[ServiceInstance]:
        marathon_apps_with_clients = get_marathon_apps_with_clients(
            clients=self.marathon_clients.get_all_clients(), embed_tasks=True
        )
        at_risk_tasks = []
        for app, client in marathon_apps_with_clients:
            for task in app.tasks:
                if task.host in draining_hosts:
                    at_risk_tasks.append(task)
        self.log.info(f"At risk tasks: {at_risk_tasks}")
        service_instances: List[ServiceInstance] = []
        for task in at_risk_tasks:
            app_id = task.app_id.strip("/")
            service, instance, _, __ = deformat_job_id(app_id)
            # check we haven't already added this instance,
            # no need to add the same instance to the bounce queue
            # more than once
            if not any(
                [
                    (service, instance) == (si.service, si.instance)
                    for si in service_instances
                ]
            ):
                # https://github.com/python/mypy/issues/2852
                service_instances.append(
                    ServiceInstance(  # type: ignore
                        service=service,
                        instance=instance,
                        cluster=self.config.get_cluster(),
                        bounce_by=int(time.time()),
                        watcher=type(self).__name__,
                        bounce_timers=None,
                        failures=0,
                    )
                )
        return service_instances


class PublicConfigEventHandler(pyinotify.ProcessEvent):
    def my_init(self, filewatcher):
        self.filewatcher = filewatcher
        self.public_config = load_system_paasta_config()
        self.marathon_clients = get_marathon_clients_from_config()

    @property
    def log(self):
        name = ".".join([__name__, type(self).__name__])
        return logging.getLogger(name)

    def filter_event(self, event):
        if event.name.endswith(".json") or event.maskname == "IN_CREATE|IN_ISDIR":
            return event

    def watch_new_folder(self, event):
        if event.maskname == "IN_CREATE|IN_ISDIR" and ".~tmp~" not in event.pathname:
            self.filewatcher.wm.add_watch(
                event.pathname, self.filewatcher.mask, rec=True
            )

    def process_default(self, event):
        self.log.debug(event)
        self.watch_new_folder(event)
        event = self.filter_event(event)
        if event:
            self.log.debug("Public config changed on disk, loading new config.")
            try:
                new_config = load_system_paasta_config()
            except ValueError:
                self.log.error("Couldn't load public config, the JSON is invalid!")
                return
            service_instances: List[Tuple[str, str]] = []
            if new_config != self.public_config:
                self.log.info(
                    "Public config has changed, now checking if it affects any services config shas."
                )
                self.public_config = new_config
                all_service_instances = get_services_for_cluster(
                    cluster=self.public_config.get_cluster(),
                    instance_type="marathon",
                    soa_dir=DEFAULT_SOA_DIR,
                )
                service_instances = get_service_instances_needing_update(
                    self.marathon_clients,
                    all_service_instances,
                    self.public_config.get_cluster(),
                )
            if service_instances:
                self.log.info(
                    f"{len(service_instances)} service instances affected. Doing a staggered bounce."
                )
                bounce_rate = self.public_config.get_deployd_big_bounce_rate()
                for service_instance in rate_limit_instances(
                    instances=service_instances,
                    cluster=self.public_config.get_cluster(),
                    number_per_minute=bounce_rate,
                    watcher_name=type(self).__name__,
                    priority=99,
                ):
                    self.filewatcher.instances_to_bounce_later.put(service_instance)


class YelpSoaEventHandler(pyinotify.ProcessEvent):
    def my_init(self, filewatcher):
        self.filewatcher = filewatcher
        self.marathon_clients = get_marathon_clients_from_config()

    @property
    def log(self):
        name = ".".join([__name__, type(self).__name__])
        return logging.getLogger(name)

    def get_service_name_from_event(self, event):
        """Get service_name from the file inotify event,
        returns None if it is not an event we're interested in"""
        starts_with = ["marathon-", "deployments.json"]
        if any([event.name.startswith(x) for x in starts_with]):
            service_name = event.path.split("/")[-1]
        elif event.name.endswith(".json") and event.path.split("/")[-1] == "secrets":
            # this is needed because we put the secrets json files in a
            # subdirectory so the service name would be "secrets" otherwise
            service_name = event.path.split("/")[-2]
        else:
            service_name = None
        return service_name

    def watch_new_folder(self, event):
        if event.maskname == "IN_CREATE|IN_ISDIR" and ".~tmp~" not in event.pathname:
            self.filewatcher.wm.add_watch(
                event.pathname, self.filewatcher.mask, rec=True
            )
            try:
                file_names = os.listdir(event.pathname)
            except OSError:
                return
            if any(["marathon-" in file_name for file_name in file_names]):
                self.log.info(f"New folder with marathon files: {event.name}.")
                self.bounce_service(event.name)

    def process_default(self, event):
        self.log.debug(event)
        self.watch_new_folder(event)
        service_name = self.get_service_name_from_event(event)
        if service_name:
            self.log.info(
                f"Looking for things to bounce for {service_name} because {event.path}/{event.name} changed."
            )
            self.bounce_service(service_name)

    def bounce_service(self, service_name):
        self.log.info(
            f"Checking if any marathon instances of {service_name} need bouncing."
        )
        instances = list_all_instances_for_service(
            service=service_name,
            clusters=[self.filewatcher.cluster],
            instance_type="marathon",
            cache=False,
        )
        self.log.debug(instances)
        service_instances = [(service_name, instance) for instance in instances]
        service_instances = get_service_instances_needing_update(
            self.marathon_clients, service_instances, self.filewatcher.cluster
        )
        for service, instance in service_instances:
            self.log.info(
                f"{service}.{instance} has a new marathon app ID. Enqueuing it to be bounced."
            )
        service_instances_to_queue = [
            # https://github.com/python/mypy/issues/2852
            ServiceInstance(  # type: ignore
                service=service,
                instance=instance,
                cluster=self.filewatcher.cluster,
                bounce_by=int(time.time()),
                watcher=type(self).__name__,
                bounce_timers=None,
                failures=0,
            )
            for service, instance in service_instances
        ]
        for service_instance in service_instances_to_queue:
            self.filewatcher.instances_to_bounce_later.put(service_instance)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
from typing import Dict
from typing import List
from typing import Tuple

import a_sync
import humanize
import isodate
from marathon.models.app import MarathonTask
from requests.exceptions import ReadTimeout

from paasta_tools import marathon_tools
from paasta_tools.autoscaling.autoscaling_service_lib import get_autoscaling_info
from paasta_tools.autoscaling.autoscaling_service_lib import ServiceAutoscalingInfo
from paasta_tools.mesos_tools import get_all_slaves_for_blacklist_whitelist
from paasta_tools.mesos_tools import get_cached_list_of_running_tasks_from_frameworks
from paasta_tools.mesos_tools import get_mesos_slaves_grouped_by_attribute
from paasta_tools.mesos_tools import select_tasks_by_id
from paasta_tools.mesos_tools import status_mesos_tasks_verbose
from paasta_tools.smartstack_tools import backend_is_up
from paasta_tools.smartstack_tools import get_backends
from paasta_tools.smartstack_tools import match_backends_and_tasks
from paasta_tools.utils import _log
from paasta_tools.utils import calculate_tail_lines
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import datetime_from_utc_to_local
from paasta_tools.utils import format_table
from paasta_tools.utils import is_under_replicated
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import NoDockerImageError
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import remove_ansi_escape_sequences

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def restart_marathon_job(service, instance, app_id, client, cluster):
    name = PaastaColors.cyan(compose_job_id(service, instance))
    _log(
        service=service,
        line="EmergencyRestart: Scaling %s down to 0 instances, then letting them scale back up"
        % (name),
        component="deploy",
        level="event",
        cluster=cluster,
        instance=instance,
    )
    client.scale_app(app_id, instances=0, force=True)


def bouncing_status_human(app_count, bounce_method):
    if app_count == 0:
        return PaastaColors.red("Disabled")
    elif app_count == 1:
        return PaastaColors.green("Configured")
    elif app_count > 1:
        return PaastaColors.yellow("Bouncing (%s)" % bounce_method)
    else:
        return PaastaColors.red("Unknown (count: %s)" % app_count)


def get_bouncing_status(service, instance, client, job_config):
    # embed_tasks=True here so that we're making the same HTTP call as the other parts of this code, so the call can be
    # cached.
    apps = marathon_tools.get_matching_appids(
        service, instance, client, embed_tasks=True
    )
    app_count = len(apps)
    bounce_method = job_config.get_bounce_method()
    return bouncing_status_human(app_count, bounce_method)


def desired_state_human(desired_state, instances):
    if desired_state == "start" and instances != 0:
        return PaastaColors.bold("Started")
    elif desired_state == "start" and instances == 0:
        return PaastaColors.bold("Stopped")
    elif desired_state == "stop":
        return PaastaColors.red("Stopped")
    else:
        return PaastaColors.red("Unknown (desired_state: %s)" % desired_state)


def status_desired_state(
    service: str,
    instance: str,
    client: marathon_tools.MarathonClient,
    job_config: marathon_tools.MarathonServiceConfig,
) -> str:
    status = get_bouncing_status(service, instance, client, job_config)
    desired_state = desired_state_human(
        job_config.get_desired_state(), job_config.get_instances()
    )
    return f"Desired State:      {status} and {desired_state}"


def status_marathon_job_human(
    service: str,
    instance: str,
    deploy_status: str,
    desired_app_id: str,
    app_count: int,
    running_instances: int,
    normal_instance_count: int,
) -> str:
    name = PaastaColors.cyan(compose_job_id(service, instance))

    if app_count >= 0:
        if running_instances >= normal_instance_count:
            status = PaastaColors.green("Healthy")
            instance_count = PaastaColors.green(
                "(%d/%d)" % (running_instances, normal_instance_count)
            )
        elif running_instances == 0:
            status = PaastaColors.yellow("Critical")
            instance_count = PaastaColors.red(
                "(%d/%d)" % (running_instances, normal_instance_count)
            )
        else:
            status = PaastaColors.yellow("Warning")
            instance_count = PaastaColors.yellow(
                "(%d/%d)" % (running_instances, normal_instance_count)
            )
        return "Marathon:   {} - up with {} instances. Status: {}".format(
            status, instance_count, deploy_status
        )
    else:
        status = PaastaColors.yellow("Warning")
        return "Marathon:   {} - {} (app {}) is not configured in Marathon yet (waiting for bounce)".format(
            status, name, desired_app_id
        )


def marathon_app_deploy_status_human(status, backoff_seconds=None):
    status_string = marathon_tools.MarathonDeployStatus.tostring(status)

    if status == marathon_tools.MarathonDeployStatus.Waiting:
        deploy_status = (
            "%s (new tasks waiting for capacity to become available)"
            % PaastaColors.red(status_string)
        )
    elif status == marathon_tools.MarathonDeployStatus.Delayed:
        deploy_status = "{} (tasks are crashing, next won't launch for another {} seconds)".format(
            PaastaColors.red(status_string), backoff_seconds
        )
    elif status == marathon_tools.MarathonDeployStatus.Deploying:
        deploy_status = PaastaColors.yellow(status_string)
    elif status == marathon_tools.MarathonDeployStatus.Stopped:
        deploy_status = PaastaColors.grey(status_string)
    elif status == marathon_tools.MarathonDeployStatus.Running:
        deploy_status = PaastaColors.bold(status_string)
    else:
        deploy_status = status_string

    return deploy_status


def status_marathon_job(
    service: str,
    instance: str,
    cluster: str,
    soa_dir: str,
    dashboards: Dict[marathon_tools.MarathonClient, str],
    normal_instance_count: int,
    clients: marathon_tools.MarathonClients,
    job_config: marathon_tools.MarathonServiceConfig,
    desired_app_id: str,
    verbose: int,
) -> Tuple[List[MarathonTask], str]:
    marathon_apps_with_clients = marathon_tools.get_marathon_apps_with_clients(
        clients=clients.get_all_clients_for_service(job_config),
        embed_tasks=True,
        service_name=service,
    )
    all_tasks = []
    all_output = [
        ""
    ]  # One entry that will be replaced with status_marathon_job_human output later.

    running_instances = 0

    if verbose > 0:
        autoscaling_info = get_autoscaling_info(marathon_apps_with_clients, job_config)
        if autoscaling_info:
            all_output.append("  Autoscaling Info:")
            headers = [
                field.replace("_", " ").capitalize()
                for field in ServiceAutoscalingInfo._fields
            ]
            table = [headers, autoscaling_info]
            all_output.append(
                "\n".join(["    %s" % line for line in format_table(table)])
            )

    deploy_status_for_desired_app = "Waiting for bounce"
    matching_apps_with_clients = marathon_tools.get_matching_apps_with_clients(
        service, instance, marathon_apps_with_clients
    )
    for app, client in matching_apps_with_clients:
        all_tasks.extend(app.tasks)
        deploy_status_for_current_app, running_instances_for_current_app, out = status_marathon_app(
            marathon_client=client,
            app=app,
            service=service,
            instance=instance,
            cluster=cluster,
            soa_dir=soa_dir,
            dashboards=dashboards,
            verbose=verbose,
        )
        if app.id.lstrip("/") == desired_app_id.lstrip("/"):
            deploy_status_for_desired_app = marathon_tools.MarathonDeployStatus.tostring(
                deploy_status_for_current_app
            )

        running_instances += running_instances_for_current_app
        all_output.append(out)

    all_output[0] = status_marathon_job_human(
        service=service,
        instance=instance,
        deploy_status=deploy_status_for_desired_app,
        desired_app_id=desired_app_id,
        app_count=len(matching_apps_with_clients),
        running_instances=running_instances,
        normal_instance_count=normal_instance_count,
    )

    return all_tasks, "\n".join(all_output)


def get_marathon_dashboard(
    client: marathon_tools.MarathonClient,
    dashboards: Dict[marathon_tools.MarathonClient, str],
    app_id: str,
) -> str:
    if dashboards is not None:
        base_url = dashboards.get(client)
        if base_url:
            url = "{}/ui/#/apps/%2F{}".format(base_url.rstrip("/"), app_id.lstrip("/"))
            return "  Marathon dashboard: %s" % PaastaColors.blue(url)
    return "  Marathon app ID: %s" % PaastaColors.bold(app_id)


def status_marathon_app(
    marathon_client: marathon_tools.MarathonClient,
    app: marathon_tools.MarathonApp,
    service: str,
    instance: str,
    cluster: str,
    soa_dir: str,
    dashboards: Dict[marathon_tools.MarathonClient, str],
    verbose: int,
) -> Tuple[int, int, str]:
    """Takes a given marathon app object and returns the details about start, times, hosts, etc"""
    output = []
    create_datetime = datetime_from_utc_to_local(isodate.parse_datetime(app.version))
    output.append(get_marathon_dashboard(marathon_client, dashboards, app.id))
    output.append(
        "    "
        + " ".join(
            [
                f"{app.tasks_running} running,",
                f"{app.tasks_healthy} healthy,",
                f"{app.tasks_staged} staged",
                f"out of {app.instances}",
            ]
        )
    )
    output.append(
        "    App created: {} ({})".format(
            str(create_datetime), humanize.naturaltime(create_datetime)
        )
    )

    deploy_status = marathon_tools.get_marathon_app_deploy_status(marathon_client, app)
    app_queue = marathon_tools.get_app_queue(marathon_client, app.id)
    unused_offers_summary = marathon_tools.summarize_unused_offers(app_queue)
    if deploy_status == marathon_tools.MarathonDeployStatus.Delayed:
        _, backoff_seconds = marathon_tools.get_app_queue_status_from_queue(app_queue)
        deploy_status_human = marathon_app_deploy_status_human(
            deploy_status, backoff_seconds
        )
    else:
        deploy_status_human = marathon_app_deploy_status_human(deploy_status)
    output.append(f"    Status: {deploy_status_human}")

    if unused_offers_summary is not None and len(unused_offers_summary) > 0:
        output.append("    Possibly stalled for:")
        output.append(
            "      ".join([f"{k}: {n} times" for k, n in unused_offers_summary.items()])
        )

    if verbose > 0:
        output.append("    Tasks:")
        rows = [
            (
                "Mesos Task ID",
                "Host deployed to",
                "Deployed at what localtime",
                "Health",
            )
        ]
        for task in app.tasks:
            local_deployed_datetime = datetime_from_utc_to_local(task.staged_at)
            if task.host is not None:
                hostname = "{}:{}".format(task.host.split(".")[0], task.ports[0])
            else:
                hostname = "Unknown"
            if not task.health_check_results:
                health_check_status = PaastaColors.grey("N/A")
            elif marathon_tools.is_task_healthy(task):
                health_check_status = PaastaColors.green("Healthy")
            else:
                health_check_status = PaastaColors.red("Unhealthy")

            rows.append(
                (
                    get_short_task_id(task.id),
                    hostname,
                    "{} ({})".format(
                        local_deployed_datetime.strftime("%Y-%m-%dT%H:%M"),
                        humanize.naturaltime(local_deployed_datetime),
                    ),
                    health_check_status,
                )
            )
        output.append("\n".join(["      %s" % line for line in format_table(rows)]))
        if len(app.tasks) == 0:
            output.append("      No tasks associated with this marathon app")
    return deploy_status, app.tasks_running, "\n".join(output)


def haproxy_backend_report(normal_instance_count, up_backends):
    """Given that a service is in smartstack, this returns a human readable
    report of the up backends"""
    # TODO: Take into account a configurable threshold, PAASTA-1102
    crit_threshold = 50
    under_replicated, ratio = is_under_replicated(
        num_available=up_backends,
        expected_count=normal_instance_count,
        crit_threshold=crit_threshold,
    )
    if under_replicated:
        status = PaastaColors.red("Critical")
        count = PaastaColors.red(
            "(%d/%d, %d%%)" % (up_backends, normal_instance_count, ratio)
        )
    else:
        status = PaastaColors.green("Healthy")
        count = PaastaColors.green("(%d/%d)" % (up_backends, normal_instance_count))
    up_string = PaastaColors.bold("UP")
    return f"{status} - in haproxy with {count} total backends {up_string} in this namespace."


def format_haproxy_backend_row(backend, is_correct_instance):
    """Pretty Prints the status of a given haproxy backend
    Takes the fields described in the CSV format of haproxy:
    http://www.haproxy.org/download/1.5/doc/configuration.txt
    And tries to make a good guess about how to represent them in text
    """
    backend_name = backend["svname"]
    backend_hostname = backend_name.split("_")[-1]
    backend_port = backend_name.split("_")[0].split(":")[-1]
    pretty_backend_name = f"{backend_hostname}:{backend_port}"
    if backend["status"] == "UP":
        status = PaastaColors.default(backend["status"])
    elif backend["status"] == "DOWN" or backend["status"] == "MAINT":
        status = PaastaColors.red(backend["status"])
    else:
        status = PaastaColors.yellow(backend["status"])
    lastcheck = "{}/{} in {}ms".format(
        backend["check_status"], backend["check_code"], backend["check_duration"]
    )
    lastchange = humanize.naturaltime(
        datetime.timedelta(seconds=int(backend["lastchg"]))
    )

    row = ("      %s" % pretty_backend_name, lastcheck, lastchange, status)

    if is_correct_instance:
        return row
    else:
        return tuple(
            PaastaColors.grey(remove_ansi_escape_sequences(col)) for col in row
        )


def status_smartstack_backends(
    service,
    instance,
    job_config,
    service_namespace_config,
    cluster,
    tasks,
    expected_count,
    soa_dir,
    synapse_port,
    synapse_haproxy_url_format,
    system_deploy_blacklist,
    system_deploy_whitelist,
    verbose,
):
    """Returns detailed information about smartstack backends for a service
    and instance.
    return: A newline separated string of the smarststack backend status
    """
    output = []

    registration = job_config.get_registrations()[0]

    discover_location_type = service_namespace_config.get_discover()
    monitoring_blacklist = job_config.get_monitoring_blacklist(
        system_deploy_blacklist=system_deploy_blacklist
    )

    filtered_slaves = get_all_slaves_for_blacklist_whitelist(
        blacklist=monitoring_blacklist, whitelist=[]
    )

    grouped_slaves = get_mesos_slaves_grouped_by_attribute(
        slaves=filtered_slaves, attribute=discover_location_type
    )

    # rebuild the dict, replacing the slave object
    # with just their hostname
    grouped_slave_hostname = {
        attribute_value: [slave["hostname"] for slave in slaves]
        for attribute_value, slaves in grouped_slaves.items()
    }

    if len(grouped_slave_hostname) == 0:
        output.append(
            "Smartstack: ERROR - %s is NOT in smartstack at all!" % registration
        )
    else:
        output.append("Smartstack:")
        if verbose:
            output.append("  Haproxy Service Name: %s" % registration)
            output.append("  Backends:")

        output.extend(
            pretty_print_smartstack_backends_for_locations(
                registration=registration,
                tasks=tasks,
                locations=grouped_slave_hostname,
                expected_count=expected_count,
                verbose=verbose,
                synapse_port=synapse_port,
                synapse_haproxy_url_format=synapse_haproxy_url_format,
            )
        )
    return "\n".join(output)


def pretty_print_smartstack_backends_for_locations(
    registration,
    tasks,
    locations,
    expected_count,
    verbose,
    synapse_port,
    synapse_haproxy_url_format,
):
    """
    Pretty prints the status of smartstack backends of a specified service and instance in the specified locations
    """
    rows = [("      Name", "LastCheck", "LastChange", "Status")] if verbose else []
    expected_count_per_location = int(expected_count / len(locations))
    for location in sorted(locations):
        hosts = locations[location]
        # arbitrarily choose the first host with a given attribute to query for replication stats
        synapse_host = hosts[0]
        sorted_backends = sorted(
            get_backends(
                registration,
                synapse_host=synapse_host,
                synapse_port=synapse_port,
                synapse_haproxy_url_format=synapse_haproxy_url_format,
            ),
            key=lambda backend: backend["status"],
            reverse=True,  # Specify reverse so that backends in 'UP' are placed above 'MAINT'
        )
        matched_tasks = match_backends_and_tasks(sorted_backends, tasks)
        running_count = sum(
            1 for backend, task in matched_tasks if backend and backend_is_up(backend)
        )
        rows.append(
            "    {} - {}".format(
                location,
                haproxy_backend_report(expected_count_per_location, running_count),
            )
        )

        # If verbose mode is specified, show status of individual backends
        if verbose:
            for backend, task in matched_tasks:
                if backend is not None:
                    rows.append(format_haproxy_backend_row(backend, task is not None))
    return format_table(rows)


def get_short_task_id(task_id):
    """Return just the Marathon-generated UUID of a Mesos task id."""
    return task_id.split(marathon_tools.MESOS_TASK_SPACER)[-1]


def status_mesos_tasks(
    service: str, instance: str, normal_instance_count: int, verbose: int
) -> str:
    job_id = marathon_tools.format_job_id(service, instance)
    # We have to add a spacer at the end to make sure we only return
    # things for service.main and not service.main_foo
    filter_string = f"{job_id}{marathon_tools.MESOS_TASK_SPACER}"

    try:
        count = len(
            select_tasks_by_id(
                a_sync.block(get_cached_list_of_running_tasks_from_frameworks),
                filter_string,
            )
        )
        if count >= normal_instance_count:
            status = PaastaColors.green("Healthy")
            count_str = PaastaColors.green("(%d/%d)" % (count, normal_instance_count))
        elif count == 0:
            status = PaastaColors.red("Critical")
            count_str = PaastaColors.red("(%d/%d)" % (count, normal_instance_count))
        else:
            status = PaastaColors.yellow("Warning")
            count_str = PaastaColors.yellow("(%d/%d)" % (count, normal_instance_count))
        running_string = PaastaColors.bold("TASK_RUNNING")
        output = (
            f"Mesos:      {status} - {count_str} tasks in the {running_string} state."
        )
    except ReadTimeout:
        return "Error: talking to Mesos timed out. It may be overloaded."

    if verbose > 0:
        tail_lines = calculate_tail_lines(verbose_level=verbose)
        output += "\n" + status_mesos_tasks_verbose(
            filter_string=filter_string,
            get_short_task_id=get_short_task_id,
            tail_lines=tail_lines,
        )

    return output


def get_marathon_dashboard_links(marathon_clients, system_paasta_config):
    """Return a dict of marathon clients and their corresponding dashboard URLs"""
    cluster = system_paasta_config.get_cluster()
    try:
        links = (
            system_paasta_config.get_dashboard_links().get(cluster).get("Marathon RO")
        )
    except KeyError:
        pass
    if isinstance(links, list) and len(links) >= len(marathon_clients.current):
        return {client: url for client, url in zip(marathon_clients.current, links)}
    return None


def perform_command(
    command: str,
    service: str,
    instance: str,
    cluster: str,
    verbose: int,
    soa_dir: str,
    clients: marathon_tools.MarathonClients,
    job_config: marathon_tools.MarathonServiceConfig,
    app_id: str = None,
) -> int:
    """Performs a start/stop/restart/status on an instance
    :param command: String of start, stop, restart, status
    :param service: service name
    :param instance: instance name, like "main" or "canary"
    :param cluster: cluster name
    :param verbose: int verbosity level
    :param client: MarathonClient or CachingMarathonClient
    :returns: A unix-style return code
    """
    system_config = load_system_paasta_config()

    if not app_id:
        try:
            app_id = job_config.format_marathon_app_dict()["id"]
        except NoDockerImageError:
            job_id = compose_job_id(service, instance)
            paasta_print(
                "Docker image for %s not in deployments.json. Exiting. Has Jenkins deployed it?"
                % job_id
            )
            return 1

    normal_instance_count = job_config.get_instances()

    current_client = clients.get_current_client_for_service(job_config)

    if command == "restart":
        restart_marathon_job(service, instance, app_id, current_client, cluster)
    elif command == "status":
        paasta_print(
            status_desired_state(service, instance, current_client, job_config)
        )
        dashboards = get_marathon_dashboard_links(clients, system_config)
        tasks, out = status_marathon_job(
            service=service,
            instance=instance,
            cluster=cluster,
            soa_dir=soa_dir,
            dashboards=dashboards,
            normal_instance_count=normal_instance_count,
            clients=clients,
            job_config=job_config,
            desired_app_id=app_id,
            verbose=verbose,
        )
        paasta_print(out)
        service_namespace_config = marathon_tools.load_service_namespace_config(
            service=service, namespace=job_config.get_nerve_namespace(), soa_dir=soa_dir
        )

        paasta_print(
            status_mesos_tasks(service, instance, normal_instance_count, verbose)
        )

        proxy_port = service_namespace_config.get("proxy_port")
        if proxy_port is not None:
            normal_smartstack_count = marathon_tools.get_expected_instance_count_for_namespace(
                service, instance, cluster
            )
            paasta_print(
                status_smartstack_backends(
                    service=service,
                    instance=instance,
                    cluster=cluster,
                    job_config=job_config,
                    service_namespace_config=service_namespace_config,
                    tasks=tasks,
                    expected_count=normal_smartstack_count,
                    soa_dir=soa_dir,
                    verbose=verbose > 0,
                    synapse_port=system_config.get_synapse_port(),
                    synapse_haproxy_url_format=system_config.get_synapse_haproxy_url_format(),
                    system_deploy_blacklist=system_config.get_deploy_blacklist(),
                    system_deploy_whitelist=system_config.get_deploy_whitelist(),
                )
            )
    else:
        # The command parser shouldn't have let us get this far...
        raise NotImplementedError("Command %s is not implemented!" % command)
    return 0


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./check_flink_services_replication.py [options]
"""
import datetime
import logging
from typing import Sequence

from paasta_tools import flink_tools
from paasta_tools import monitoring_tools
from paasta_tools.check_services_replication_tools import main
from paasta_tools.flink_tools import FlinkDeploymentConfig
from paasta_tools.kubernetes_tools import filter_pods_by_service_instance
from paasta_tools.kubernetes_tools import is_pod_ready
from paasta_tools.kubernetes_tools import V1Pod
from paasta_tools.smartstack_tools import KubeSmartstackReplicationChecker


log = logging.getLogger(__name__)


def container_lifetime(pod: V1Pod,) -> datetime.timedelta:
    """Return a time duration for how long the pod is alive
    """
    st = pod.status.start_time
    return datetime.datetime.now(st.tzinfo) - st


def healthy_flink_containers_cnt(si_pods: Sequence[V1Pod], container_type: str) -> int:
    """Return count of healthy Flink containers with given type
    """
    return len(
        [
            pod
            for pod in si_pods
            if pod.metadata.labels["flink-container-type"] == container_type
            and is_pod_ready(pod)
            and container_lifetime(pod).total_seconds() > 60
        ]
    )


def check_flink_service_replication(
    instance_config: FlinkDeploymentConfig,
    all_pods: Sequence[V1Pod],
    smartstack_replication_checker: KubeSmartstackReplicationChecker,
) -> None:
    si_pods = filter_pods_by_service_instance(
        pod_list=all_pods,
        service=instance_config.service,
        instance=instance_config.instance,
    )
    taskmanagers_expected_cnt = instance_config.config_dict.get(
        "taskmanager", {"instances": 10}
    ).get("instances", 10)
    num_healthy_supervisors = healthy_flink_containers_cnt(si_pods, "supervisor")
    num_healthy_jobmanagers = healthy_flink_containers_cnt(si_pods, "jobmanager")
    num_healthy_taskmanagers = healthy_flink_containers_cnt(si_pods, "taskmanager")

    # TBD: check cnt according to Flink

    monitoring_tools.send_replication_event_if_under_replication(
        instance_config=instance_config,
        expected_count=1,
        num_available=num_healthy_supervisors,
        sub_component="supervisor",
    )
    monitoring_tools.send_replication_event_if_under_replication(
        instance_config=instance_config,
        expected_count=1,
        num_available=num_healthy_jobmanagers,
        sub_component="jobmanager",
    )
    monitoring_tools.send_replication_event_if_under_replication(
        instance_config=instance_config,
        expected_count=taskmanagers_expected_cnt,
        num_available=num_healthy_taskmanagers,
        sub_component="taskmanager",
    )


if __name__ == "__main__":
    main(
        flink_tools.FlinkDeploymentConfig,
        check_flink_service_replication,
        namespace="paasta-flinks",
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import os
import re

import a_sync

from . import exceptions
from . import framework
from . import mesos_file
from paasta_tools.async_utils import async_ttl_cache


class Task:

    cmd_re = re.compile(r"\(Command: (.+)\)")

    def __init__(self, master, items):
        self.master = master
        self.__items = items

    def __str__(self):
        return "{}:{}".format(a_sync.block(self.slave), self["id"])

    def __getitem__(self, name):
        return self.__items[name]

    async def executor(self):
        return await (await self.slave()).task_executor(self["id"])

    async def framework(self):
        return framework.Framework(await self.master.framework(self["framework_id"]))

    @async_ttl_cache(cleanup_self=True)
    async def directory(self):
        try:
            return (await self.executor())["directory"]
        except exceptions.MissingExecutor:
            return ""

    @async_ttl_cache(cleanup_self=True)
    async def slave(self):
        return await self.master.slave(self["slave_id"])

    async def file(self, path):
        return mesos_file.File(await self.slave(), self, path)

    async def file_list(self, path):
        return await (await self.slave()).file_list(os.path.join(self.directory, path))

    async def stats(self):
        try:
            return await (await self.slave()).task_stats(self["id"])
        except exceptions.MissingExecutor:
            return {}

    async def cpu_time(self):
        st = await self.stats()
        secs = st.get("cpus_user_time_secs", 0) + st.get("cpus_system_time_secs", 0)
        # timedelta has a resolution of .000000 while mesos only keeps .00
        return str(datetime.timedelta(seconds=secs)).rsplit(".", 1)[0]

    async def cpu_limit(self):
        return (await self.stats()).get("cpus_limit", 0)

    async def mem_limit(self):
        return (await self.stats()).get("mem_limit_bytes", 0)

    async def rss(self):
        return (await self.stats()).get("mem_rss_bytes", 0)

    async def command(self):
        try:
            result = self.cmd_re.search((await self.executor())["name"])
        except exceptions.MissingExecutor:
            result = None
        if not result:
            return "none"
        return result.group(1)

    async def user(self):
        return (await self.framework()).user



#!/usr/bin/env python
from paasta_tools import chronos_tools

if __name__ == "__main__":
    config = chronos_tools.load_chronos_config()
    client = chronos_tools.get_chronos_client(config)
    jobs = [job["name"] for job in client.list()]

    for job in jobs:
        client.delete(job)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains methods used by the paasta client to generate a Jenkins build
pipeline."""
import re

from paasta_tools.cli.utils import guess_service_name
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import NoSuchService
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.monitoring_tools import get_team
from paasta_tools.monitoring_tools import get_team_email_address
from paasta_tools.utils import _log_audit
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_git_url
from paasta_tools.utils import list_services
from paasta_tools.utils import paasta_print


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        "generate-pipeline",
        help="Configures a Yelp-specific Jenkins build pipeline to match the 'deploy.yaml'",
        description=(
            "'paasta generate-pipeline' is a Yelp-specific tool to interact with Jenkins "
            "to build a build pipeline that matches what is declared in the 'deploy.yaml' "
            "for a service."
        ),
        epilog="Warning: Due to the Yelpisms in this tool, it is not currently useful to other organizations.",
    )
    list_parser.add_argument(
        "-s",
        "--service",
        help="Name of service for which you wish to generate a Jenkins pipeline",
    ).completer = lazy_choices_completer(list_services)
    list_parser.set_defaults(command=paasta_generate_pipeline)


def paasta_generate_pipeline(args):
    """Generate a Jenkins build pipeline.
    :param args: argparse.Namespace obj created from sys.args by cli"""
    service = args.service or guess_service_name()
    soa_dir = DEFAULT_SOA_DIR
    try:
        validate_service_name(service, soa_dir=soa_dir)
    except NoSuchService as service_not_found:
        paasta_print(service_not_found)
        return 1

    generate_pipeline(service=service, soa_dir=soa_dir)
    _log_audit(action="generate-pipeline", service=service)


def get_git_repo_for_fab_repo(service, soa_dir):
    """Returns the 'repo' in fab_repo terms. fab_repo just wants the trailing
    section of the git_url, after the colon.
    """
    git_url = get_git_url(service, soa_dir=soa_dir)
    repo = git_url.split(":")[1]
    return repo


def print_warning():
    paasta_print("Warning: paasta generate-pipeline is DEPRECATED")
    paasta_print()
    paasta_print("Please go to y/jenkinsfile to move to a jenkinsfile")
    paasta_print()
    paasta_print(
        "If you really need to run this command, manually run the commands below:"
    )
    paasta_print()
    paasta_print("Warning: running this tool on an existing pipeline will remove any")
    paasta_print("hand-made customizations and will leave behind orphaned Jenkins jobs")
    paasta_print("that need to be manually cleaned up.")
    paasta_print()
    input("Press enter to continue or ctrl-c to cancel")


def generate_pipeline(service, soa_dir):
    email_address = get_team_email_address(service=service, soa_dir=soa_dir)
    repo = get_git_repo_for_fab_repo(service, soa_dir)
    if not email_address:
        owner = get_team(overrides={}, service=service, soa_dir=soa_dir)
    else:
        # fab_repo tacks on the domain, so we only want the first
        # part of the email.
        owner = re.sub("@.*", "", email_address)
    cmds = [
        "fab_repo setup_jenkins:services/%s,"
        "profile=paasta,job_disabled=False,owner=%s,repo=%s" % (service, owner, repo),
        "fab_repo setup_jenkins:services/%s,"
        "profile=paasta_boilerplate,owner=%s,repo=%s" % (service, owner, repo),
    ]
    print_warning()
    for cmd in cmds:
        paasta_print(f"Please run: `{cmd}`")



import logging
from abc import ABC
from abc import abstractmethod
from typing import Optional

from kubernetes.client import V1DeleteOptions
from kubernetes.client import V1Deployment
from kubernetes.client.rest import ApiException

from paasta_tools.kubernetes_tools import get_deployment_config
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import KubeDeployment
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.utils import SystemPaastaConfig


class Application(ABC):
    def __init__(self, item: V1Deployment, logging=logging.getLogger(__name__)) -> None:
        self.kube_deployment = KubeDeployment(
            service=item.metadata.labels["yelp.com/paasta_service"],
            instance=item.metadata.labels["yelp.com/paasta_instance"],
            git_sha=item.metadata.labels["yelp.com/paasta_git_sha"],
            config_sha=item.metadata.labels["yelp.com/paasta_config_sha"],
            replicas=item.spec.replicas,
        )
        self.item = item
        self.soa_config = None
        self.logging = logging

    def load_local_config(
        self, soa_dir: str, system_paasta_config: SystemPaastaConfig
    ) -> Optional[KubernetesDeploymentConfig]:
        if not self.soa_config:
            self.soa_config = get_deployment_config(  # type: ignore
                self.item, soa_dir, system_paasta_config.get_cluster()
            )
        return self.soa_config

    def __str__(self):
        service = self.kube_deployment.service
        instance = self.kube_deployment.instance
        git_sha = self.kube_deployment.git_sha
        config_sha = self.kube_deployment.config_sha
        return f"{service}-{instance}-{git_sha}-{config_sha}"

    @abstractmethod
    def deep_delete(self, kube_client: KubeClient) -> None:
        """
        Remove all controllers, pods, and pod disruption budgets related to this application
        :param kube_client:
        """
        pass

    def delete_pod_disruption_budget(self, kube_client: KubeClient) -> None:
        try:
            kube_client.policy.delete_namespaced_pod_disruption_budget(
                name=self.item.metadata.name,
                namespace=self.item.metadata.namespace,
                body=V1DeleteOptions(),
            )
        except ApiException as e:
            if e.status == 404:
                # Deployment does not exist, nothing to delete but
                # we can consider this a success.
                self.logging.debug(
                    "not deleting nonexistent pod disruption budget/{} from namespace/{}".format(
                        self.item.metadata.name, self.item.metadata.namespace
                    )
                )
            else:
                raise e
        else:
            self.logging.info(
                "deleted pod disruption budget/{} from namespace/{}".format(
                    self.item.metadata.name, self.item.metadata.namespace
                )
            )


class DeploymentWrapper(Application):
    def deep_delete(self, kube_client: KubeClient) -> None:
        """
        Remove all controllers, pods, and pod disruption budgets related to this application
        :param kube_client:
        """
        delete_options = V1DeleteOptions(propagation_policy="Foreground")
        try:
            kube_client.deployments.delete_namespaced_deployment(
                self.item.metadata.name, self.item.metadata.namespace, delete_options
            )
        except ApiException as e:
            if e.status == 404:
                # Deployment does not exist, nothing to delete but
                # we can consider this a success.
                self.logging.debug(
                    "not deleting nonexistent deploy/{} from namespace/{}".format(
                        self.item.metadata.name, self.item.metadata.namespace
                    )
                )
            else:
                raise e
        else:
            self.logging.info(
                "deleted deploy/{} from namespace/{}".format(
                    self.item.metadata.name, self.item.metadata.namespace
                )
            )
        self.delete_pod_disruption_budget(kube_client)


class StatefulSetWrapper(Application):
    def deep_delete(self, kube_client: KubeClient) -> None:
        """
        Remove all controllers, pods, and pod disruption budgets related to this application
        :param kube_client:
        """
        delete_options = V1DeleteOptions(propagation_policy="Foreground")
        try:
            kube_client.deployments.delete_namespaced_stateful_set(
                self.item.metadata.name, self.item.metadata.namespace, delete_options
            )
        except ApiException as e:
            if e.status == 404:
                # StatefulSet does not exist, nothing to delete but
                # we can consider this a success.
                self.logging.debug(
                    "not deleting nonexistent statefulset/{} from namespace/{}".format(
                        self.item.metadata.name, self.item.metadata.namespace
                    )
                )
            else:
                raise e
        else:
            self.logging.info(
                "deleted statefulset/{} from namespace/{}".format(
                    self.item.metadata.name, self.item.metadata.namespace
                )
            )
        self.delete_pod_disruption_budget(kube_client)



import logging
import time
from collections import namedtuple
from queue import PriorityQueue
from queue import Queue
from threading import Thread
from typing import Any
from typing import Collection
from typing import List
from typing import Optional
from typing import Tuple

from paasta_tools.marathon_tools import DEFAULT_SOA_DIR
from paasta_tools.marathon_tools import get_all_marathon_apps
from paasta_tools.marathon_tools import get_marathon_clients
from paasta_tools.marathon_tools import get_marathon_servers
from paasta_tools.marathon_tools import load_marathon_service_config
from paasta_tools.marathon_tools import load_marathon_service_config_no_cache
from paasta_tools.marathon_tools import MarathonClients
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import NoDeploymentsAvailable
from paasta_tools.utils import NoDockerImageError

BounceTimers = namedtuple(
    "BounceTimers", ["processed_by_worker", "setup_marathon", "bounce_length"]
)
BaseServiceInstance = namedtuple(
    "ServiceInstance",
    [
        "service",
        "instance",
        "bounce_by",
        "watcher",
        "bounce_timers",
        "failures",
        "priority",
        "processed_count",
    ],
)


class ServiceInstance(BaseServiceInstance):
    __slots__ = ()

    def __new__(
        _cls,
        service: str,
        instance: str,
        watcher: str,
        cluster: str,
        bounce_by: float,
        failures: int = 0,
        bounce_timers: Optional[BounceTimers] = None,
        priority: Optional[int] = None,
        processed_count: int = 0,
    ) -> "ServiceInstance":
        if priority is None:
            priority = get_priority(service, instance, cluster)
        return super().__new__(  # type: ignore
            _cls=_cls,
            service=service,
            instance=instance,
            watcher=watcher,
            bounce_by=bounce_by,
            failures=failures,
            bounce_timers=bounce_timers,
            priority=priority,
            processed_count=processed_count,
        )


def get_priority(service: str, instance: str, cluster: str) -> int:
    try:
        config = load_marathon_service_config(
            service=service, instance=instance, cluster=cluster, soa_dir=DEFAULT_SOA_DIR
        )
    except (
        NoDockerImageError,
        InvalidJobNameError,
        NoDeploymentsAvailable,
        NoConfigurationForServiceError,
    ):
        return 0
    return config.get_bounce_priority()


class PaastaThread(Thread):
    @property
    def log(self) -> logging.Logger:
        name = ".".join([type(self).__module__, type(self).__name__])
        return logging.getLogger(name)


class PaastaQueue(Queue):
    def __init__(self, name: str, *args: Any, **kwargs: Any) -> None:
        self.name = name
        super().__init__(*args, **kwargs)

    @property
    def log(self) -> logging.Logger:
        name = ".".join([type(self).__module__, type(self).__name__])
        return logging.getLogger(name)

    def put(self, item: Any, *args: Any, **kwargs: Any) -> None:
        self.log.debug(f"Adding {item} to {self.name} queue")
        super().put(item, *args, **kwargs)


class PaastaPriorityQueue(PriorityQueue):
    def __init__(self, name: str, *args: Any, **kwargs: Any) -> None:
        self.name = name
        super().__init__(*args, **kwargs)
        self.counter = 0

    @property
    def log(self) -> logging.Logger:
        name = ".".join([type(self).__module__, type(self).__name__])
        return logging.getLogger(name)

    def put_with_priority(
        self, priority: float, item: Any, *args: Any, **kwargs: Any
    ) -> None:
        self.log.debug(f"Adding {item} to {self.name} queue with priority {priority}")
        # this counter is to preserve the FIFO nature of the queue, it increments on every put
        # and the python PriorityQueue sorts based on the first item in the tuple (priority)
        # and then the second item in the tuple (counter). This way all items with the same
        # priority come out in the order they were entered.
        self.counter += 1
        super().put((priority, self.counter, item), *args, **kwargs)

    def get(self, *args: Any, **kwargs: Any) -> Any:
        return super().get(*args, **kwargs)[2]


def rate_limit_instances(
    instances: Collection[Tuple[str, str]],
    cluster: str,
    number_per_minute: float,
    watcher_name: str,
    priority: Optional[float] = None,
) -> List[ServiceInstance]:
    service_instances = []
    if not instances:
        return []
    time_now = int(time.time())
    time_step = int(60 / number_per_minute)
    bounce_time = time_now
    for service, instance in instances:
        # https://github.com/python/mypy/issues/2852
        service_instances.append(
            ServiceInstance(  # type: ignore
                service=service,
                instance=instance,
                watcher=watcher_name,
                cluster=cluster,
                bounce_by=bounce_time,
                bounce_timers=None,
                failures=0,
                priority=priority,
            )
        )
        bounce_time += time_step
    return service_instances


def exponential_back_off(
    failures: int, factor: float, base: float, max_time: float
) -> float:
    seconds = factor * base ** failures
    return seconds if seconds < max_time else max_time


def get_service_instances_needing_update(
    marathon_clients: MarathonClients,
    instances: Collection[Tuple[str, str]],
    cluster: str,
) -> List[Tuple[str, str]]:
    marathon_apps = {}
    for marathon_client in marathon_clients.get_all_clients():
        marathon_apps.update(
            {app.id: app for app in get_all_marathon_apps(marathon_client)}
        )

    marathon_app_ids = marathon_apps.keys()
    service_instances = []
    for service, instance in instances:
        try:
            config = load_marathon_service_config_no_cache(
                service=service,
                instance=instance,
                cluster=cluster,
                soa_dir=DEFAULT_SOA_DIR,
            )
            config_app = config.format_marathon_app_dict()
            app_id = "/{}".format(config_app["id"])
        # Not ideal but we rely on a lot of user input to create the app dict
        # and we really can't afford to bail if just one app definition is malformed
        except Exception as e:
            print(
                "ERROR: Skipping {}.{} because: '{}'".format(service, instance, str(e))
            )
            continue
        if app_id not in marathon_app_ids:
            service_instances.append((service, instance))
        elif marathon_apps[app_id].instances != config_app["instances"]:
            service_instances.append((service, instance))
    return service_instances


def get_marathon_clients_from_config() -> MarathonClients:
    system_paasta_config = load_system_paasta_config()
    marathon_servers = get_marathon_servers(system_paasta_config)
    marathon_clients = get_marathon_clients(marathon_servers)
    return marathon_clients



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import sys

from a_sync import block

from paasta_tools.mesos.exceptions import MasterNotAvailableException
from paasta_tools.mesos_tools import get_mesos_master
from paasta_tools.metrics.metastatus_lib import assert_no_duplicate_frameworks
from paasta_tools.utils import paasta_print


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--check",
        "-C",
        dest="check",
        type=str,
        default="",
        help="Comma separated list of frameworks to check for duplicates",
    )
    return parser.parse_args()


def check_mesos_no_duplicate_frameworks() -> None:
    options = parse_args()
    check = options.check.split(",")
    master = get_mesos_master()
    try:
        state = block(master.state)
    except MasterNotAvailableException as e:
        paasta_print("CRITICAL: %s" % e.args[0])
        sys.exit(2)

    result = assert_no_duplicate_frameworks(state, check)
    if result.healthy:
        paasta_print("OK: " + result.message)
        sys.exit(0)
    else:
        paasta_print(result.message)
        sys.exit(2)


if __name__ == "__main__":
    check_mesos_no_duplicate_frameworks()



#!/usr/bin/env python
import datetime
import inspect
import logging
import socket
import sys
import time
from queue import Empty

import humanize
import service_configuration_lib

from paasta_tools.deployd import watchers
from paasta_tools.deployd.common import get_marathon_clients_from_config
from paasta_tools.deployd.common import PaastaPriorityQueue
from paasta_tools.deployd.common import PaastaQueue
from paasta_tools.deployd.common import PaastaThread
from paasta_tools.deployd.common import rate_limit_instances
from paasta_tools.deployd.common import ServiceInstance
from paasta_tools.deployd.leader import PaastaLeaderElection
from paasta_tools.deployd.metrics import QueueMetrics
from paasta_tools.deployd.workers import PaastaDeployWorker
from paasta_tools.list_marathon_service_instances import (
    get_service_instances_that_need_bouncing,
)
from paasta_tools.marathon_tools import DEFAULT_SOA_DIR
from paasta_tools.metrics.metrics_lib import get_metrics_interface
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import ZookeeperPool


class DedupedPriorityQueue(PaastaPriorityQueue):
    """This class extends the python Queue class so that the Queue is
    deduplicated. i.e. there can be only one copy of each service instance
    in the queue at any one time
    """

    def __init__(self, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.bouncing = set()

    def put_with_priority(self, priority, service_instance, *args, **kwargs):
        service_instance_key = "{}.{}".format(
            service_instance.service, service_instance.instance
        )
        if service_instance_key not in self.bouncing:
            self.bouncing.add(service_instance_key)
            super().put_with_priority(priority, service_instance, *args, **kwargs)
        else:
            self.log.debug(
                f"{service_instance_key} already present in {self.name}, dropping extra message"
            )

    def get(self, *args, **kwargs):
        service_instance = super().get(*args, **kwargs)
        service_instance_key = "{}.{}".format(
            service_instance.service, service_instance.instance
        )
        self.bouncing.remove(service_instance_key)
        return service_instance


class Inbox(PaastaThread):
    def __init__(self, instances_to_bounce_later, instances_to_bounce_now):
        super().__init__()
        self.daemon = True
        self.name = "Inbox"
        self.instances_to_bounce_later = instances_to_bounce_later
        self.instances_to_bounce_now = instances_to_bounce_now
        self.to_bounce = {}

    def run(self):
        while True:
            self.process_inbox()

    def process_inbox(self):
        """Takes things from the bounce_later queue, adds them to the to_bounce queue, then
        takes things from the to_bounce queue and puts them on the bounce_now queue when they are due"""
        try:
            service_instance = self.instances_to_bounce_later.get(block=False)
        except Empty:
            service_instance = None
        if service_instance:
            self.log.debug(
                f"Processing {service_instance.service}.{service_instance.instance} from the bounce_later queue to see if we need to bounce it now"
            )  # noqa: E501
            self.process_service_instance(service_instance)
        if self.instances_to_bounce_later.empty() and self.to_bounce:
            self.process_to_bounce()
        time.sleep(0.1)

    def process_service_instance(self, service_instance):
        service_instance_key = f"{service_instance.service}.{service_instance.instance}"
        if self.should_add_to_bounce(service_instance, service_instance_key):
            self.to_bounce[service_instance_key] = service_instance

    def should_add_to_bounce(self, service_instance, service_instance_key):
        if service_instance_key in self.to_bounce:
            if (
                service_instance.bounce_by
                > self.to_bounce[service_instance_key].bounce_by
            ):
                self.log.debug(
                    f"{service_instance} already in to_bounce queue with higher priority"
                )
                return False
        return True

    def process_to_bounce(self):
        bounced = []
        self.log.debug(
            f"Processing {len(self.to_bounce.keys())} to_bounce queue entries..."
        )
        for service_instance_key in self.to_bounce.keys():
            if self.to_bounce[service_instance_key].bounce_by < int(time.time()):
                service_instance = self.to_bounce[service_instance_key]
                human_bounce_by = humanize.naturaldelta(
                    datetime.timedelta(
                        seconds=(time.time() - service_instance.bounce_by)
                    )
                )
                bounced.append(service_instance_key)
                self.log.info(
                    f"Enqueuing {service_instance.service}.{service_instance.instance} to the bounce_now queue (bounce_by {human_bounce_by} ago)"
                )  # noqa E501
                self.instances_to_bounce_now.put_with_priority(
                    service_instance.priority, service_instance
                )
        for service_instance_key in bounced:
            self.to_bounce.pop(service_instance_key)
        # TODO: if the bounceq is empty we could probably start adding SIs from
        # self.to_bounce to make sure the workers always have something to do.


class AddHostnameFilter(logging.Filter):
    def __init__(self):
        super().__init__()
        self.hostname = socket.gethostname()

    def filter(self, record):
        record.hostname = self.hostname
        return True


class DeployDaemon(PaastaThread):
    def __init__(self):
        super().__init__()
        self.started = False
        self.daemon = True
        service_configuration_lib.disable_yaml_cache()
        self.config = load_system_paasta_config()
        self.setup_logging()
        self.metrics = get_metrics_interface("paasta.deployd")
        self.instances_to_bounce_now = DedupedPriorityQueue("instances_to_bounce_now")
        self.instances_to_bounce_later = PaastaQueue(
            "instances_to_bounce_later"
        )  # noqa: E501
        self.control = PaastaQueue("ControlQueue")
        self.inbox = Inbox(self.instances_to_bounce_later, self.instances_to_bounce_now)
        self.marathon_clients = get_marathon_clients_from_config()

    def setup_logging(self):
        root_logger = logging.getLogger()
        root_logger.setLevel(getattr(logging, self.config.get_deployd_log_level()))
        handler = logging.StreamHandler()
        handler.addFilter(AddHostnameFilter())
        root_logger.addHandler(handler)
        logging.getLogger("kazoo").setLevel(logging.CRITICAL)
        handler.setFormatter(
            logging.Formatter(
                "%(asctime)s:%(hostname)s:%(levelname)s:%(name)s:%(message)s"
            )
        )

    def run(self):
        self.log.info("paasta-deployd starting up...")
        startup_counter = self.metrics.create_counter(
            "process_started", paasta_cluster=self.config.get_cluster()
        )
        startup_counter.count()
        with ZookeeperPool() as self.zk:
            self.election = PaastaLeaderElection(
                self.zk,
                "/paasta-deployd-leader",
                socket.getfqdn(),
                control=self.control,
            )
            self.is_leader = False
            self.log.info("Waiting to become leader")
            self.election.run(self.startup)
            self.log.info("Leadership given up, exiting...")

    def bounce(self, service_instance):
        self.instances_to_bounce_later.put(service_instance)

    @property
    def watcher_threads_enabled(self):
        disabled_watchers = self.config.get_disabled_watchers()
        watcher_classes = [
            obj[1]
            for obj in inspect.getmembers(watchers)
            if inspect.isclass(obj[1]) and obj[1].__bases__[0] == watchers.PaastaWatcher
        ]
        enabled_watchers = [
            x for x in watcher_classes if x.__name__ not in disabled_watchers
        ]
        return enabled_watchers

    def startup(self):
        self.is_leader = True
        self.log.info("This node is elected as leader {}".format(socket.getfqdn()))
        leader_counter = self.metrics.create_counter(
            "leader_elections", paasta_cluster=self.config.get_cluster()
        )
        leader_counter.count()
        QueueMetrics(
            inbox=self.inbox,
            cluster=self.config.get_cluster(),
            metrics_provider=self.metrics,
        ).start()
        self.inbox.start()
        self.log.info("Starting all watcher threads")
        self.start_watchers()
        self.log.info(
            "All watchers started, now adding all services for initial bounce"
        )
        self.add_all_services()
        self.log.info("Prioritising services that we know need a bounce...")
        if self.config.get_deployd_startup_oracle_enabled():
            self.prioritise_bouncing_services()
        self.log.info("Starting worker threads")
        self.start_workers()
        self.started = True
        self.log.info("Startup finished!")
        self.main_loop()

    def main_loop(self):
        while True:
            try:
                message = self.control.get(block=False)
            except Empty:
                message = None
            if message == "ABORT":
                self.log.info("Got ABORT message, main_loop exiting...")
                break
            if not self.all_watchers_running():
                self.log.error("One or more watcher died, committing suicide!")
                sys.exit(1)
            if self.all_workers_dead():
                self.log.error("All workers have died, committing suicide!")
                sys.exit(1)
            self.check_and_start_workers()
            time.sleep(0.1)

    def all_watchers_running(self):
        return all([watcher.is_alive() for watcher in self.watcher_threads])

    def all_workers_dead(self):
        return all([not worker.is_alive() for worker in self.workers])

    def check_and_start_workers(self):
        live_workers = len([worker for worker in self.workers if worker.is_alive()])
        number_of_dead_workers = self.config.get_deployd_number_workers() - live_workers
        for i in range(number_of_dead_workers):
            self.log.error("Detected a dead worker, starting a replacement thread")
            worker_no = len(self.workers) + 1
            worker = PaastaDeployWorker(
                worker_no,
                self.instances_to_bounce_later,
                self.instances_to_bounce_now,
                self.config,
                self.metrics,
            )
            worker.start()
            self.workers.append(worker)

    def stop(self):
        self.control.put("ABORT")

    def start_workers(self):
        self.workers = []
        for i in range(self.config.get_deployd_number_workers()):
            worker = PaastaDeployWorker(
                i,
                self.instances_to_bounce_later,
                self.instances_to_bounce_now,
                self.config,
                self.metrics,
            )
            worker.start()
            self.workers.append(worker)

    def add_all_services(self):
        instances = get_services_for_cluster(
            cluster=self.config.get_cluster(),
            instance_type="marathon",
            soa_dir=DEFAULT_SOA_DIR,
        )
        instances_to_add = rate_limit_instances(
            instances=instances,
            cluster=self.config.get_cluster(),
            number_per_minute=self.config.get_deployd_startup_bounce_rate(),
            watcher_name="daemon_start",
            priority=99,
        )
        for service_instance in instances_to_add:
            self.instances_to_bounce_later.put(service_instance)

    def prioritise_bouncing_services(self):
        service_instances = get_service_instances_that_need_bouncing(
            self.marathon_clients, DEFAULT_SOA_DIR
        )
        for service_instance in service_instances:
            self.log.info(f"Prioritising {service_instance} to be bounced immediately")
            service, instance = service_instance.split(".")
            self.instances_to_bounce_later.put(
                ServiceInstance(
                    service=service,
                    instance=instance,
                    cluster=self.config.get_cluster(),
                    watcher=type(self).__name__,
                    bounce_by=int(time.time()),
                    bounce_timers=None,
                    failures=0,
                )
            )

    def start_watchers(self):
        """ should block until all threads happy"""
        self.watcher_threads = [
            watcher(
                instances_to_bounce_later=self.instances_to_bounce_later,
                cluster=self.config.get_cluster(),
                zookeeper_client=self.zk,
                config=self.config,
            )
            for watcher in self.watcher_threads_enabled
        ]

        self.log.info(f"Starting the following watchers {self.watcher_threads}")
        for watcher in self.watcher_threads:
            watcher.start()
        self.log.info("Waiting for all watchers to start")
        attempts = 0
        while attempts < 120:
            if all([watcher.is_ready for watcher in self.watcher_threads]):
                return
            self.log.info("Sleeping and waiting for watchers to all start")
            self.log.info(
                "Waiting on: {}".format(
                    [
                        watcher.__class__.__name__
                        for watcher in self.watcher_threads
                        if not watcher.is_ready
                    ]
                )
            )
            time.sleep(1)
            attempts += 1
        self.log.error("Failed to start all the watchers, exiting...")
        sys.exit(1)


def main():
    dd = DeployDaemon()
    dd.start()
    while dd.is_alive():
        time.sleep(0.1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse

from paasta_tools import tron_tools
from paasta_tools.utils import paasta_print


def parse_args():
    parser = argparse.ArgumentParser(
        description="Lists Tron namespaces for a cluster, excluding MASTER"
    )
    parser.add_argument(
        "-c",
        "--cluster",
        dest="cluster",
        default=None,
        help="Use a different Tron cluster",
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        default=tron_tools.DEFAULT_SOA_DIR,
        help="Use a different soa config directory",
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    namespaces = tron_tools.get_tron_namespaces(
        cluster=args.cluster, soa_dir=args.soa_dir
    )
    paasta_print("\n".join(namespaces))


if __name__ == "__main__":
    main()
